/* orinoco.h
 * 
 * Common definitions to all pieces of the various orinoco
 * drivers
 */

#ifndef _ORINOCO_H
#define _ORINOCO_H

#include <linux/types.h>
#include <linux/spinlock.h>
#include <linux/netdevice.h>
#include <linux/wireless.h>
#include <linux/version.h>

/* for mem_dev */
#include <linux/devfs_fs_kernel.h>

#include "hermes.h"

/* Workqueue / task queue backwards compatibility stuff */

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,41)
#include <linux/workqueue.h>
#else
#include <linux/tqueue.h>
#define work_struct tq_struct
#define INIT_WORK INIT_TQUEUE
#define schedule_work schedule_task
#endif

/* To enable debug messages */
//#define ORINOCO_DEBUG		3

#if (! defined (WIRELESS_EXT)) || (WIRELESS_EXT < 10)
#error "orinoco driver requires Wireless extensions v10 or later."
#endif /* (! defined (WIRELESS_EXT)) || (WIRELESS_EXT < 10) */
#define WIRELESS_SPY		// enable iwspy support

#define ORINOCO_MAX_KEY_SIZE	14
#define ORINOCO_MAX_KEYS	4

#define ORINOCO_DUMMY_FID	0xFFFF
#define ORINOCO_TXFIDS		4

#define ORINOCO_PREPARE_RAMDL	1
#define	ORINOCO_FINISH_RAMDL	2

#define ORINOCO_AP_ENABLED 	1
#define	ORINOCO_AP_HIDDEN	2
#define ORINOCO_AP_LEAVE	4

struct wdslink { 
        struct net_device dev;
	u16 info;
	u8 peer_addr[ETH_ALEN];
	u16 rts_thres;
	u16 bitratemode;
        struct net_device_stats stats;
	struct iw_statistics wstats;
};      
 
#define WDS_PORTS    	6
#define WDS_INFO_OPEN	1

struct orinoco_key {
	u16 len;	/* always stored as little-endian */
	char data[ORINOCO_MAX_KEY_SIZE];
} __attribute__ ((packed));

#define ORINOCO_INTEN	 	( HERMES_EV_RX | HERMES_EV_ALLOC | HERMES_EV_TX | \
				HERMES_EV_TXEXC | HERMES_EV_WTERR | HERMES_EV_INFO | \
				HERMES_EV_INFDROP )


struct orinoco_private {
	void *card;	/* Pointer to card dependant structure */
	int (*hard_reset)(struct orinoco_private *);

	/* Synchronisation stuff */
	spinlock_t lock;
	int hw_unavailable;
	struct work_struct timeout_task;

	/* driver state */
	int open;
	u16 last_linkstatus;
	int connected;
	int reset_active;

	/* Net device stuff */
	struct net_device *ndev;
	struct net_device_stats stats;
	struct iw_statistics wstats;

	/* Hardware control variables */
	hermes_t hw;
	u16 txfid[ORINOCO_TXFIDS];
	int txfids;

	/* Capabilities of the hardware/firmware */
	int firmware_type;
#define FIRMWARE_TYPE_AGERE 1
#define FIRMWARE_TYPE_INTERSIL 2
#define FIRMWARE_TYPE_SYMBOL 3
	int has_ibss, has_port3, has_ibss_any, ibss_port;
	int has_wep, has_big_wep;
	int has_mwo;
	int has_pm;
	int has_preamble;
	int has_sensitivity;
	int nicbuf_size;
	u16 channel_mask;

	/* Configuration paramaters */
	u32 iw_mode;
	int prefer_port3;
	u16 wep_on, wep_restrict, tx_key;
	struct orinoco_key keys[ORINOCO_MAX_KEYS];
	int bitratemode;
 	char nick[IW_ESSID_MAX_SIZE+1];
	char desired_essid[IW_ESSID_MAX_SIZE+1];
	u16 frag_thresh, mwo_robust;
	u16 channel;
	u16 ap_density, rts_thresh;
	u16 pm_on, pm_mcast, pm_period, pm_timeout;
	u16 preamble;
#ifdef WIRELESS_SPY
	int			spy_number;
	u_char			spy_address[IW_MAX_SPY][ETH_ALEN];
	struct iw_quality	spy_stat[IW_MAX_SPY];
#endif

	/* Configuration dependent variables */
	int port_type, createibss;
	int promiscuous, mc_count;
	
	/* mem_dev stuff */
	devfs_handle_t mem_dev;
	int mem_open;
	pid_t dl_pid;

	/* BSS master mode */
	u32 ap_mode;
	struct wdslink wdslink[WDS_PORTS];
};

#ifdef ORINOCO_DEBUG
extern int orinoco_debug;
#define DEBUG(n, args...) do { if (orinoco_debug>(n)) printk(KERN_DEBUG args); } while(0)
#else
#define DEBUG(n, args...) do { } while (0)
#endif	/* ORINOCO_DEBUG */

#define TRACE_ENTER(devname) DEBUG(2, "%s: -> " __FUNCTION__ "()\n", devname);
#define TRACE_EXIT(devname)  DEBUG(2, "%s: <- " __FUNCTION__ "()\n", devname);

extern struct net_device *alloc_orinocodev(int sizeof_card,
					   int (*hard_reset)(struct orinoco_private *));
extern int __orinoco_up(struct net_device *dev);
extern int __orinoco_down(struct net_device *dev);
int orinoco_reinit_firmware(struct net_device *dev);

extern void orinoco_interrupt(int irq, void * dev_id, struct pt_regs *regs);

/* mem_dev stuff */
extern int orinoco_mem_dev_init(struct orinoco_private *priv);
extern void orinoco_mem_dev_cleanup(struct orinoco_private *priv);

#define ORINOCO_REGISTER	0
#define ORINOCO_UNREGISTER	1
#define ORINOCO_ATTACH		2
#define ORINOCO_DETACH		3
#define ORINOCO_ACTION_MASK	7
#define ORINOCO_REMOVED		0x80
int orinoco_action(struct orinoco_private *priv, int action);

/********************************************************************/
/* Locking and synchronization functions                            */
/********************************************************************/

/* These functions *must* be inline or they will break horribly on
 * SPARC, due to its weird semantics for save/restore flags. extern
 * inline should prevent the kernel from linking or module from
 * loading if they are not inlined. */
extern inline int orinoco_lock(struct orinoco_private *priv,
			       unsigned long *flags)
{
	spin_lock_irqsave(&priv->lock, *flags);
	if (priv->hw_unavailable) {
		printk(KERN_DEBUG "orinoco_lock() called with hw_unavailable (dev=%p)\n",
		       priv->ndev);
		spin_unlock_irqrestore(&priv->lock, *flags);
		return -EBUSY;
	}
	return 0;
}

extern inline void orinoco_unlock(struct orinoco_private *priv,
				  unsigned long *flags)
{
	spin_unlock_irqrestore(&priv->lock, *flags);
}

extern inline u16 orinoco_pop_txfid(struct orinoco_private *priv) {

	if(!priv->txfids) 
		return ORINOCO_DUMMY_FID;
	else
		return priv->txfid[--(priv->txfids)];
}

extern inline int orinoco_push_txfid(struct orinoco_private *priv, u16 fid) {

	if(priv->txfids < ORINOCO_TXFIDS)
		priv->txfid[(priv->txfids)++] = fid;
	else
		return -EINVAL;

	return 0;
}

extern inline int orinoco_lookup_port(struct net_device *dev) {
	struct orinoco_private *priv = dev->priv;
	int ret;
	
	if(dev == priv->ndev)
		return 0;
	for(ret=1;ret<=WDS_PORTS;ret++) {
		if(dev == &(priv->wdslink[ret-1].dev))
			return ret;
	}
	
	return -EINVAL;
}

extern inline int orinoco_wds_active(struct orinoco_private *priv) {
	int count;
	int res=0;
	
	for(count=0;count<WDS_PORTS;count++) {
		if((priv->wdslink[count].info&WDS_INFO_OPEN) && (priv->wdslink[count].dev.name[0]))
			res++;
	}
	
	return res;
}

#endif /* _ORINOCO_H */
