/*
 * functions for hfwload
 * (firmware loader for hermes wavelan cards)
 *
 * Copyright (c) 2003, Benedikt 'Hunz' Heinz <hermesap@hunz.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 */
 
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "hfw.h"
#include "pdar.h"

int hex2bin(char *in, unsigned char *out, int max) {
	int res=0;
	
	while((res>>1)<max) {
		if(!(res&1))
			out[res>>1]=0;
		if((in[res]>='A') && (in[res]<='F'))
			out[res>>1]|=(in[res]-'A'+10)<<((res&1) ? 0 : 4);
		else if((in[res]>='a') && (in[res]<='f'))
			out[res>>1]|=(in[res]-'a'+10)<<((res&1) ? 0 : 4);
		else if((in[res]>='0') && (in[res]<='9'))
			out[res>>1]|=(in[res]-'0')<<((res&1) ? 0 : 4);
		else
			return (res>>1);
		res++;
	}
	return max;
}

int plug_data(struct segment *segm,unsigned long offset, unsigned long length, unsigned char *pd) {
	int len;
	
	for(;segm->length;segm++) {
		len=segm->length;
		if(segm->data_len>len)
			len=segm->data_len;
		if((offset>=segm->offset) && (offset+length<=segm->offset+len)) {
			//printf("plug %8lX %d\n",offset-segm->offset,length);
			memcpy(segm->data+(offset-segm->offset),pd,length);
			return 0;
		}
	}
	return -1;
}

int hfw_plug_records(struct hfw *hfw, unsigned char *opda, unsigned char *pda, int verbose) {
	struct plugrecord *pr;
	unsigned char *pd;
		
	if((!pda) || (!hfw) || (!(pr=hfw->plugs)))
		return -1;
	for(;pr->length;pr++) {
		pd=NULL;
		if((pr->code&0xFFF0)==0xFFF0) {
			/* special plug */
			/* TODO!! */
		}
		else if((pd=lookup_pda(pr->code&0xFFFF,pr->length,opda))) {
			if(verbose)
				fprintf(stderr,"overriding PDA entry for plugrecord %08lX %8lX\n",pr->code,pr->length);
		}
		else if(!(pd=lookup_pda(pr->code&0xFFFF,pr->length,pda))) {
			if (verbose)
				fprintf(stderr,"no matching PDA entry for plugrecord %08lX %8lX\n",pr->code,pr->length);
		}
		if(pd) {
			/* plug */
			if(plug_data(hfw->segms,pr->offset,pr->length,pd)) {
				fprintf(stderr,"segment for plugrecord %08lX %8lX %8lX not found\n",pr->code,pr->offset,pr->length);
				return -1;
			}
		} /* pd */
	} /* foreach plugrecord loop */
	return 0;
}

int write_hfw_plugs(FILE *fl, struct plugrecord *recs) {
	int count;
	
	if(!recs)
		return 0;
	for(count=0; recs[count].code && recs[count].length; count++)
		fprintf(fl, "PLUG  %08lX %08lX %08lX\n", recs[count].code, recs[count].offset/*-0x393800+0x33000*/, recs[count].length);
	
	return count;
}

void free_hfw(struct hfw *hfw) {
	int count;
	
	free(hfw->plugs);
	if(hfw->segms) {
		for(count=0;hfw->segms[count].length;count++)
			free(hfw->segms[count].data);
		free(hfw->segms);
	}
	free(hfw);
}

int add_plug(struct hfw *hfw, unsigned long code, unsigned long offset, unsigned long length) {
	struct plugrecord *pr=hfw->plugs;
	int count=1;
	
	if(!length)
		return -1;
	if(!pr) {
		if(!(pr=malloc(sizeof(struct plugrecord))))
			return -1;
		memset(pr,0,sizeof(struct plugrecord));
		hfw->plugs=pr;
	}
	else {
		while(pr[count-1].length)
			count++;
	}
	if(!(pr=realloc(hfw->plugs,(count+1)*sizeof(struct plugrecord))))
		return -1;
	hfw->plugs=pr;
	memset(pr+count,0,sizeof(struct plugrecord));
	
	pr[count-1].length=length;
	pr[count-1].offset=offset;
	pr[count-1].code=code;
	return 0;
}

struct segment *add_segm(struct hfw *hfw, unsigned long offset, unsigned long length, unsigned long unknown) {
	struct segment *sr=hfw->segms;
	int count=1;
	
	if(!sr) {
		if(!(sr=malloc(sizeof(struct segment))))
			return NULL;
		memset(sr,0,sizeof(struct segment));
		hfw->segms=sr;
	}
	else {
		while(sr[count-1].length)
			count++;
	}
	if(!(sr=realloc(hfw->segms,(count+1)*sizeof(struct segment))))
		return NULL;
	hfw->segms=sr;
	memset(sr+count,0,sizeof(struct segment));
	
	if(!(sr[count-1].data=malloc(length)))
		return NULL;
	sr[count-1].length=length;
	sr[count-1].unknown=unknown;
	sr[count-1].offset=offset;
	return sr+count-1;
}

inline int add_data(struct segment *segm, unsigned char *data) {
	if((!segm) || (!segm->length) || (!segm->data))
		return -1;
	if((segm->data_len+2)>segm->length) {
		unsigned char *db;
		
		if(!(db=realloc(segm->data,segm->length+128))) {
			fprintf(stderr,"can't expand segment length %ld by 128 bytes\n",segm->length);
			return -1;
		}
		segm->data=db;
		segm->length+=128;
	}
	segm->data[segm->data_len++]=data[0];
	segm->data[segm->data_len++]=data[1];
	return 0;
}

void dump_segments(struct segment *sr) {

	for(;(sr) && (sr->length);sr++)
		printf("segm ofs %8lX len %8lX dlen %4X unk %8lX\n",sr->offset,sr->length,sr->data_len,sr->unknown);
}

struct hfw *read_hfw(unsigned char *file) {
	unsigned long args[3],clen=0;
	unsigned char buf[128], *dp, data[2];
	int count, len;
	FILE *fl;
	struct hfw *ret;
	struct segment *csegm=NULL;
	
	if(!(fl=fopen(file,"rt")))
		return NULL;
	if((!fgets(buf,sizeof(buf),fl)) || (strcmp(buf,"HFW1\n"))) {
		fprintf(stderr,"read_hfw: not an HFW1-file\n");
		return NULL;
	}
	if(!(ret=malloc(sizeof(struct hfw))))
		return NULL;
	memset(ret,0,sizeof(struct hfw));
	while(fgets(buf,sizeof(buf),fl)) {
		len=strlen(buf);
		if(len>=1) {
			if(buf[len-1]!=0x0a) {
				free_hfw(ret);
				return NULL;
			}
			else {
				buf[--len]=0;
				if((len>=1) && (buf[len-1]==0x0d))
					buf[--len]=0;
			}
		}
		if(!strncmp(buf,"#",1)) {
			// comment
		}
		else if(!strncmp(buf,"ENTRY ",6)) {
			// entrypoint
			dp=buf+5;
			SKIP_SPACES(dp);
			if(sscanf(dp,"%lx",&(ret->entry))!=1) {
				fprintf(stderr,"read_hfw: line invalid: %s\n",buf);
				free_hfw(ret);
				return NULL;
			}
		}
		else if(!strncmp(buf,"PLUG ",5)) {
			// pluginfo
			dp=buf+4;
			for(count=0;count<3;count++) {
				SKIP_SPACES(dp);
				if(sscanf(dp,"%08lx",args+count)!=1) {
					fprintf(stderr,"read_hfw: line invalid: %s\n",buf);
					free_hfw(ret);
					return NULL;
				}
				dp+=8;
			}
			if(add_plug(ret,args[0],args[1],args[2])) {
				fprintf(stderr,"read_hfw: adding plugrecord %08lX %08lX %08lX failed\n",args[0],args[1],args[2]);
				free_hfw(ret);
				return NULL;
			}
		}
		else if(!strncmp(buf,"SEG ",4)) {
			// segment
			dp=buf+3;
			for(count=0;count<3;count++) {
				SKIP_SPACES(dp);
				if(sscanf(dp,"%08lx",args+count)!=1) {
					fprintf(stderr,"read_hfw: line invalid: %s\n",buf);
					free_hfw(ret);
					return NULL;
				}
				dp+=8;
			}
			if((clen>0) && (clen<csegm->length)) {
				if(!(dp=realloc(csegm->data,csegm->data_len))) {
					fprintf(stderr,"read_hfw: error shrinking segment data buffer\n");
					dump_segments(csegm);
					free_hfw(ret);
					return NULL;
				}
				else
					csegm->data=dp;
				fprintf(stderr,"segment size raised from %8lX to %4X\n",clen,csegm->data_len);
				csegm->length=clen;
			}
			if(!(csegm=add_segm(ret,args[0],args[1],args[2]))) {
				fprintf(stderr,"read_hfw: adding segmrecord %08lX %08lX %08lX failed\n",args[0],args[1],args[2]);
				free_hfw(ret);
				return NULL;
			}
			clen=csegm->length;
		}
		else if(!strncmp(buf,"DATA ",5)) {
			// data
			dp=buf+4;
			while(1) {
				SKIP_SPACES(dp);
				if(!dp[0])
					break;
				if(hex2bin(dp,data,2)!=2) {
					fprintf(stderr,"read_hfw: line invalid: %s\n",buf);
					free_hfw(ret);
					return NULL;
				}
				dp+=4;
				if(add_data(csegm,data)) {
					dump_segments(csegm);
					fprintf(stderr,"read_hfw: error adding data line %s\n",buf);
					free_hfw(ret);
					return NULL;
				}
			}
		}
		else {
			// garbage
			fprintf(stderr,"read_hfw: line invalid: %s\n",buf);
			free_hfw(ret);
			return NULL;
		}
	}
	return ret;
}

int hfw_upload(struct hfw *hfw, int fd) {
	struct segment *sr;
	int count,count2,res;
	unsigned char *ptr;
	
#ifdef TEST
	fd=2;
#endif
	if(!hfw)
		return -1;
	for(sr=hfw->segms;sr->length;sr++) {
//		printf("%8lX\n",sr->offset);
#ifndef TEST
		if(lseek(fd,sr->offset,SEEK_SET)!=sr->offset) {
			fprintf(stderr,"seeking to offset %8lX failed\n",sr->offset);
			return -1;
		}
#endif
		count=sr->data_len;
		ptr=sr->data;
		while(count) {
//			printf("%lX\n",lseek(fd,0,SEEK_CUR));
			if(count>=512)
				count2=512;
			else
				count2=count;
			if((res=write(fd,ptr,count2))<=0) {
				fprintf(stderr,"write error: %d\n",res);
				return -1;
			}
			count-=count2;
			ptr+=count2;
		}
	}
	return 0;
}
